import collections.abc

try:
    from graphviz import Digraph
except ImportError:
    pass
from relion._parser.processnode import ProcessNode


class ProcessGraph(collections.abc.Sequence):
    def __init__(self, node_list):
        self._node_list = node_list

    def __eq__(self, other):
        if isinstance(other, ProcessGraph):
            if len(self) == len(other):
                for n in self:
                    if n not in other:
                        return False
                return True
        return False

    def __hash__(self):
        return hash(("relion._parser.pipeline.ProcessGraph", iter(self._node_list)))

    def __iter__(self):
        return iter(self._node_list)

    def __len__(self):
        return len(self._node_list)

    def __getitem__(self, index):
        if not isinstance(index, int):
            raise ValueError("Index of ProcessGraph must be an integer")
        return self._node_list[index]

    def extend(self, other):
        if not isinstance(other, ProcessGraph):
            raise ValueError("Can only extend a ProcessGraph with another ProcessGraph")
        self._node_list.extend(other._node_list)

    def index(self, node):
        return self._node_list.index(node)

    def link_from_to(self, from_node, to_node):
        self[self.index(from_node)].link_to(to_node)

    def node_explore(self, node, explored):
        if not isinstance(node, ProcessNode):
            raise ValueError(
                "ProcessGraph.node_explore must be called with a ProcessNode as the starting point; a string or similar is insufficient"
            )
        if node not in explored:
            explored.append(node)
        for next_node in node:
            self.node_explore(next_node, explored)

    def add_node(self, new_node):
        if isinstance(new_node, ProcessNode):
            self._node_list.append(new_node)
        else:
            raise ValueError("Attempted to add a node that was not a ProcessNode")

    def remove_node(self, node_name):
        behind_nodes = []
        for currnode in self:
            if node_name in currnode:
                behind_nodes.append(currnode)
                currnode.unlink_from(node_name)
        for bnode in behind_nodes:
            for next_node in self._node_list[self._node_list.index(node_name)]:
                bnode.link_to(next_node)
        self._node_list.remove(node_name)

    def find_origins(self):
        child_nodes = []
        for node in self:
            child_nodes.extend([next_node for next_node in node])
        origins = [p for p in self if p not in child_nodes]
        return origins

    def merge(self, other):
        node_names = [p._path for p in self]
        other_names = [p._path for p in other]
        if len(set(node_names).intersection(set(other_names))) > 0:
            for new_node in other:
                if new_node._path not in self:
                    self.add_node(new_node)
                else:
                    for next_node in new_node:
                        if next_node._path not in self[self.index(new_node._path)]:
                            self[self.index(new_node._path)].link_to(next_node)
            return True
        else:
            return False

    def split_connected(self):
        connected_graphs = []
        for origin in self.find_origins():
            curr_graph = []
            self.node_explore(origin, curr_graph)
            connected_graphs.append(ProcessGraph(curr_graph))
        for_removal = []
        for i, g in enumerate(connected_graphs):
            for ng in connected_graphs[i + 1 :]:
                if g.merge(ng):
                    if ng not in for_removal:
                        for_removal.append(ng)
        for r in for_removal:
            connected_graphs.remove(r)
        return connected_graphs

    def _split_connected(self, connected_dict, origin, origins_dict):
        connected_graphs = self.split_connected()
        ancillary_count = 1
        for cg in connected_graphs:
            if origin in cg:
                connected_dict["main"] = cg
                origins_dict["main"] = self[self.index(origin)]
            else:
                connected_dict[f"ancillary:{ancillary_count}"] = cg
                origins_dict[f"ancillary:{ancillary_count}"] = cg.find_origins()[0]
                ancillary_count += 1

    def show_all_nodes(self):
        try:
            digraph = Digraph(format="svg")
            digraph.attr(rankdir="LR")
            for node in self:
                digraph.node(name=str(node._path))
                for next_node in node:
                    digraph.edge(str(node._path), str(next_node._path))
            digraph.render("pipeline.gv")
        except Exception:
            raise Warning(
                "Failed to create nodes display. Your environment may not have graphviz available."
            )

    def wipe(self):
        self._node_list = []



from gemmi import cif
import os
import pathlib
import warnings

try:
    from graphviz import Digraph
except ImportError:
    pass
import copy
import datetime
import calendar
from relion._parser.processnode import ProcessNode
from relion._parser.processgraph import ProcessGraph


class RelionPipeline:
    def __init__(self, origin, graphin=ProcessGraph([])):
        self.origin = origin
        self._nodes = graphin
        self._connected = {}
        self.origins = {}
        self._job_nodes = ProcessGraph([])
        self._jobtype_nodes = ProcessGraph([])
        self._connected_jobs = {}
        self.job_origins = {}
        self._jobs_collapsed = False

    def __iter__(self):
        if not self._jobs_collapsed:
            self._collapse_jobs_to_jobtypes()
        return iter(self._jobtype_nodes)

    @staticmethod
    def _request_star_values(star_path, column, search=None):
        if search is None:
            search = column
        gemmi_readable_path = os.fspath(star_path)
        star_doc = cif.read_file(gemmi_readable_path)
        for block_index, block in enumerate(star_doc):
            if list(block.find_loop(search)):
                block_number = block_index
        data_block = star_doc[block_number]
        values = data_block.find_loop(column)
        return list(values)

    def _load_file_nodes_from_star(self, star_path):
        return ProcessGraph(
            [
                ProcessNode(pathlib.Path(p))
                for p in self._request_star_values(star_path, "_rlnPipeLineNodeName")
            ]
        )

    def _load_job_nodes_from_star(self, star_path):
        return ProcessGraph(
            [
                ProcessNode(pathlib.Path(p), alias=al)
                for p, al in zip(
                    self._request_star_values(star_path, "_rlnPipeLineProcessName"),
                    self._request_star_values(star_path, "_rlnPipeLineProcessAlias"),
                )
            ]
        )

    def load_nodes_from_star(self, star_path):
        self._nodes.wipe()
        file_nodes = self._load_file_nodes_from_star(star_path)
        job_nodes = self._load_job_nodes_from_star(star_path)
        self._nodes.extend(file_nodes)
        self._nodes.extend(job_nodes)
        binding_pairs = [
            (ProcessNode(p1), ProcessNode(p2))
            for p1, p2 in zip(
                self._request_star_values(star_path, "_rlnPipeLineEdgeFromNode"),
                self._request_star_values(
                    star_path,
                    "_rlnPipeLineEdgeProcess",
                    search="_rlnPipeLineEdgeFromNode",
                ),
            )
        ]
        binding_pairs.extend(
            [
                (ProcessNode(p1), ProcessNode(p2))
                for p1, p2 in zip(
                    self._request_star_values(
                        star_path,
                        "_rlnPipeLineEdgeProcess",
                        search="_rlnPipeLineEdgeToNode",
                    ),
                    self._request_star_values(star_path, "_rlnPipeLineEdgeToNode"),
                )
            ]
        )
        for f, t in binding_pairs:
            self._nodes[self._nodes.index(f._path)].link_to(
                self._nodes[self._nodes.index(t._path)]
            )
        self._nodes._split_connected(self._connected, self.origin, self.origins)
        self._set_job_nodes(star_path)

    def check_job_node_statuses(self, basepath):
        for node in self._job_nodes:
            success = basepath / node._path / "RELION_JOB_EXIT_SUCCESS"
            failure = basepath / node._path / "RELION_JOB_EXIT_FAILURE"
            aborted = basepath / node._path / "RELION_JOB_EXIT_ABORTED"
            if failure.is_file():
                node.attributes["status"] = False
                node.attributes["end_time_stamp"] = datetime.datetime.fromtimestamp(
                    failure.stat().st_ctime
                )
            elif aborted.is_file():
                node.attributes["status"] = False
                node.attributes["end_time_stamp"] = datetime.datetime.fromtimestamp(
                    aborted.stat().st_ctime
                )
            elif success.is_file():
                node.attributes["status"] = True
                node.attributes["end_time_stamp"] = datetime.datetime.fromtimestamp(
                    success.stat().st_ctime
                )
            else:
                node.attributes["status"] = None

    def _set_job_nodes(self, star_path):
        self._job_nodes = copy.deepcopy(self._nodes)
        file_nodes = self._load_file_nodes_from_star(star_path)
        for fnode in file_nodes:
            self._job_nodes.remove_node(fnode._path)
        self._job_nodes._split_connected(
            self._connected_jobs, self.origin, self.job_origins
        )

    def _collapse_jobs_to_jobtypes(self):
        ordered_graph = []
        self._job_nodes.node_explore(
            self._job_nodes[self._job_nodes.index(self.origin)], ordered_graph
        )
        self._jobtype_nodes = ProcessGraph(copy.deepcopy(ordered_graph))
        for node in self._jobtype_nodes:
            node.attributes["job"] = node._path.name
            node._path = node._path.parent
        self._jobs_collapsed = True

    def show_all_nodes(self):
        self._nodes.show_all_nodes()

    def show_job_nodes(self, basepath):
        try:
            digraph = Digraph(format="svg")
        except Exception:
            warnings.warn(
                "Failed to create nodes display. Your environment may not have graphviz avaliable."
            )
            return
        digraph.attr(rankdir="LR")
        running_node = self.current_job
        for node in self._job_nodes:
            if node._path == running_node:
                bordercolour = "teal"
            else:
                bordercolour = "purple"
            if node.attributes["status"]:
                node_colour = "mediumseagreen"
            elif node.attributes["status"] is None:
                node_colour = "lightgray"
            else:
                node_colour = "orangered"
            if node.attributes["start_time_stamp"] is not None:
                sstamp = node.attributes["start_time_stamp"]
            else:
                sstamp = "???"
            if node.attributes["end_time_stamp"] is not None:
                estamp = node.attributes["end_time_stamp"]
            else:
                estamp = "???"
            if (
                node.attributes.get("alias") is not None
                and node.attributes.get("alias") != "None"
            ):
                nodename = node.attributes.get("alias")
            else:
                nodename = str(node._path)
            digraph.node(
                name=nodename,
                shape="hexagon",
                style="filled",
                fillcolor=node_colour,
                color=bordercolour,
                tooltip=f"Start: {sstamp}&#013;End: {estamp}&#013;Path: {node._path}",
            )
            for next_node in node:
                stime = next_node.attributes["start_time"]
                etime = next_node.attributes["end_time"]
                jcount = next_node.attributes["job_count"]

                if (
                    next_node.attributes.get("alias") is not None
                    and next_node.attributes.get("alias") != "None"
                ):
                    next_nodename = next_node.attributes.get("alias")
                else:
                    next_nodename = str(next_node._path)

                if stime is not None and etime is not None and jcount != 1:
                    digraph.edge(
                        nodename,
                        next_nodename,
                        label=f"{stime} / {etime} [{jcount}]",
                    )
                elif stime is not None and etime is not None:
                    digraph.edge(
                        nodename,
                        next_nodename,
                        label=f"{stime} / {etime}",
                    )
                elif stime is not None:
                    digraph.edge(
                        nodename,
                        next_nodename,
                        label=f"{stime} / ???",
                    )
                elif etime is not None:
                    digraph.edge(
                        nodename,
                        next_nodename,
                        label=f"??? / {etime}",
                    )
                else:
                    digraph.edge(nodename, next_nodename, label="??? / ???")
        digraph.render(basepath / "Pipeline" / "relion_pipeline_jobs.gv")

    def collect_job_times(self, schedule_logs):
        for job in self._job_nodes:
            jtime, jcount = self._lookup_job_time(schedule_logs, job)
            job.attributes["start_time_stamp"] = jtime
            job.attributes["job_count"] = jcount
        self._calculate_relative_job_times()

    def _calculate_relative_job_times(self):
        times = [j.attributes["start_time_stamp"] for j in self._job_nodes]
        etimes = [j.attributes["end_time_stamp"] for j in self._job_nodes]
        already_started_times = [t for t in times if t is not None]
        origin = sorted(already_started_times)[0]
        relative_times = [(t - origin).total_seconds() for t in times if t is not None]
        just_seconds = lambda tdelta: tdelta - datetime.timedelta(
            microseconds=tdelta.microseconds
        )
        for node, rt in zip(self._job_nodes, relative_times):
            node.attributes["start_time"] = just_seconds(datetime.timedelta(seconds=rt))
        relative_etimes = [
            (t - origin).total_seconds() for t in etimes if t is not None
        ]
        for node, rt in zip(self._job_nodes, relative_etimes):
            node.attributes["end_time"] = just_seconds(datetime.timedelta(seconds=rt))

    def _lookup_job_time(self, schedule_logs, job):
        time = None
        jobcount = 0
        for log in schedule_logs:
            with open(log, "r") as slf:
                lines = slf.readlines()
                for lindex, line in enumerate(lines):
                    if "Executing" in line and str(job._path) in line:
                        split_line = lines[lindex - 1].split()
                        time_split = split_line[4].split(":")
                        dtime = datetime.datetime(
                            year=int(split_line[5]),
                            month=list(calendar.month_abbr).index(split_line[2]),
                            day=int(split_line[3]),
                            hour=int(time_split[0]),
                            minute=int(time_split[1]),
                            second=int(time_split[2]),
                        )
                        if time is None or dtime < time:
                            time = dtime
                        jobcount += 1
        return time, jobcount

    @property
    def current_jobs(self):
        running_jobs = []
        for node in self._job_nodes:
            if (
                node.attributes["start_time_stamp"] is not None
                and node.attributes["status"] is None
            ):
                running_jobs.append(node)
        if len(running_jobs) == 0:
            return None
        else:
            return running_jobs



from relion import Project
import argparse
import pathlib


def run():
    parser = argparse.ArgumentParser()
    parser.add_argument("proj_path")
    args = parser.parse_args()
    relion_dir = pathlib.Path(args.proj_path)
    proj = Project(relion_dir)
    current_jobs = proj.current_jobs
    print("These jobs are currently running: \n")
    for current_job in current_jobs:
        alias = current_job.attributes.get("alias")
        if alias is not None:
            print(f"Current job: {current_job._path} [alias={alias}]")
        else:
            print(f"Current job: {current_job._path}")
        print(f"Job started at: {current_job.attributes['start_time_stamp']}")
        print(f"Job has been run {current_job.attributes['job_count']} time(s)")
        print()


if __name__ == "__main__":
    run()



#!/usr/bin/env python3
"""
External job for making a mask de novo with a soft edge
"""

import argparse
import os
import os.path
import subprocess


RELION_JOB_FAILURE_FILENAME = "RELION_JOB_EXIT_FAILURE"
RELION_JOB_SUCCESS_FILENAME = "RELION_JOB_EXIT_SUCCESS"


def run_job(project_dir, job_dir, args_list):
    parser = argparse.ArgumentParser()
    parser.add_argument("--box_size", dest="box_size")
    parser.add_argument("--angpix", dest="angpix")
    parser.add_argument("--outer_radius", dest="outer_radius")
    parser.add_argument(
        "--j", dest="threads", help="Number of threads to run (ignored)"
    )
    args = parser.parse_args(args_list)

    os.chdir(job_dir)

    # need to remove the hard coding of outer_radius and width_soft_edge

    command = [
        "relion_mask_create",
        "--denovo",
        "true",
        "--box_size",
        f"{args.box_size}",
        "--angpix",
        f"{args.angpix}",
        "--outer_radius",
        f"{args.outer_radius}",
        "--width_soft_edge",
        "5",
    ]

    subprocess.run(command, check=True)


def main():
    """Change to the job working directory, then call run_job()"""
    parser = argparse.ArgumentParser()
    parser.add_argument("--out_dir", dest="out_dir", help="Output directory name")
    parser.add_argument(
        "--pipeline_control", help="Directory for pipeline control files"
    )
    known_args, other_args = parser.parse_known_args()
    project_dir = os.getcwd()
    os.makedirs(known_args.out_dir, exist_ok=True)
    os.chdir(known_args.out_dir)
    if os.path.isfile(RELION_JOB_FAILURE_FILENAME):
        print(" mask_soft_edge_external_job: Removing previous failure indicator file")
        os.remove(RELION_JOB_FAILURE_FILENAME)
    if os.path.isfile(RELION_JOB_SUCCESS_FILENAME):
        print(" mask_soft_edge_external_job: Removing previous success indicator file")
        os.remove(RELION_JOB_SUCCESS_FILENAME)
    try:
        os.chdir("../..")
        run_job(project_dir, known_args.out_dir, other_args)
    except Exception:
        if os.getcwd() == project_dir:
            os.chdir(known_args.out_dir)
        open(RELION_JOB_FAILURE_FILENAME, "w").close()
        raise
    else:
        if os.getcwd() == project_dir:
            os.chdir(known_args.out_dir)
        open(RELION_JOB_SUCCESS_FILENAME, "w").close()


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
"""
External job for fitting FSC curves and finding the value at which they cross 0.5
"""

import argparse
import json
import os
import os.path

import gemmi


RELION_JOB_FAILURE_FILENAME = "RELION_JOB_EXIT_FAILURE"
RELION_JOB_SUCCESS_FILENAME = "RELION_JOB_EXIT_SUCCESS"


def run_job(project_dir, out_dir, fscs_files, args_list):
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--j", dest="threads", help="Number of threads to run (ignored)"
    )

    resolutions = []

    for i, starin in enumerate(fscs_files):
        if not starin.endswith(".star"):
            raise ValueError("Input files containing FSC curves must end with .star")
        fsc_in = gemmi.cif.read_file(os.path.join(project_dir, starin))
        data_as_dict = json.loads(fsc_in.as_json())["fsc"]
        invres = [1 / x for x in data_as_dict["_rlnangstromresolution"]]
        fsc = data_as_dict["_rlnfouriershellcorrelationcorrected"]
        res = lin_interp(invres, fsc)
        resolutions.append(res)

    class_index = resolutions.index(min(resolutions))

    return class_index


def lin_interp(invres, fsc):
    cpoints = crossing_points(invres, fsc)
    last_crossing_point = cpoints[-1]
    slope = (last_crossing_point[1][1] - last_crossing_point[0][1]) / (
        last_crossing_point[1][0] - last_crossing_point[0][0]
    )
    constant = last_crossing_point[0][1] - slope * last_crossing_point[0][0]
    return (0.5 - constant) / slope


def crossing_points(invres, fsc):
    points = []
    for i in range(1, len(invres)):
        if fsc[i] <= 0.5 and fsc[i - 1] >= 0.5:
            points.append(((invres[i - 1], fsc[i - 1]), (invres[i], fsc[i])))
    return points


def main():
    """Change to the job working directory, then call run_job()"""
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--out_dir", dest="out_dir", help="Directory for the FSC fitting External job"
    )
    parser.add_argument(
        "--i",
        nargs="+",
        dest="fscs",
        help="Input star file names containing FSC curves for comparison",
    )
    parser.add_argument(
        "--pipeline_control", help="Directory for pipeline control files"
    )
    known_args, other_args = parser.parse_known_args()
    project_dir = os.getcwd()
    os.makedirs(known_args.out_dir, exist_ok=True)
    os.chdir(known_args.out_dir)
    if os.path.isfile(RELION_JOB_FAILURE_FILENAME):
        print(" fsc_fitting_external_job: Removing previous failure indicator file")
        os.remove(RELION_JOB_FAILURE_FILENAME)
    if os.path.isfile(RELION_JOB_SUCCESS_FILENAME):
        print(" fsc_fitting_external_job: Removing previous success indicator file")
        os.remove(RELION_JOB_SUCCESS_FILENAME)
    try:
        os.chdir("../..")
        class_index = run_job(
            project_dir, known_args.out_dir, known_args.fscs, other_args
        )
        with open(
            os.path.join(project_dir, known_args.out_dir, "BestClass.txt"), "w"
        ) as f:
            f.write(str(class_index))
    except Exception:
        os.chdir(known_args.out_dir)
        open(RELION_JOB_FAILURE_FILENAME, "w").close()
        raise
    else:
        os.chdir(known_args.out_dir)
        open(RELION_JOB_SUCCESS_FILENAME, "w").close()


if __name__ == "__main__":
    main()



import pathlib
import functools


@functools.total_ordering
class ProcessNode:
    def __init__(self, path, **kwargs):
        self._path = pathlib.PurePosixPath(path)
        self._out = []
        self.attributes = {}
        for key, value in kwargs.items():
            self.attributes[key] = value
        self.attributes["status"] = kwargs.get("status")
        self.attributes["start_time_stamp"] = kwargs.get("start_time_stamp")
        self.attributes["end_time_stamp"] = kwargs.get("end_time_stamp")
        self.attributes["start_time"] = kwargs.get("start_time")
        self.attributes["end_time"] = kwargs.get("end_time")

    def __eq__(self, other):
        if isinstance(other, ProcessNode):
            if self._path == other._path and len(self._out) == len(other._out):
                for n in self._out:
                    if n not in other._out:
                        return False
                return True
        else:
            try:
                return str(self._path) == str(other)
            except Exception:
                return False
        return False

    def __hash__(self):
        return hash(("relion._parser.pipeline.ProcessNode", self._path))

    def __repr__(self):
        return f"Node({repr(str(self._path))})"

    def __iter__(self):
        return iter(self._out)

    def __len__(self):
        return len(self._out)

    def __lt__(self, other):
        if self._is_child(other):
            return True
        return False

    @property
    def name(self):
        return str(self._path)

    def change_name(self, new_name):
        self._path = new_name

    def link_to(self, next_node):
        if next_node not in self._out:
            self._out.append(next_node)

    def unlink_from(self, next_node):
        if next_node in self._out:
            self._out.remove(next_node)

    def _is_child_checker(self, possible_child, checks):
        if self == possible_child:
            checks.extend([True])
        for child in self:
            checks.extend(child._is_child_checker(possible_child, checks=checks))
        return checks

    def _is_child(self, possible_child):
        if True in self._is_child_checker(possible_child, checks=[]):
            return True
        else:
            return False



from collections import namedtuple
from collections import Counter
from operator import attrgetter
from relion._parser.jobtype import JobType

Class2DParticleClass = namedtuple(
    "Class2DParticleClass",
    [
        "particle_sum",
        "reference_image",
        "class_distribution",
        "accuracy_rotations",
        "accuracy_translations_angst",
        "estimated_resolution",
        "overall_fourier_completeness",
    ],
)

Class2DParticleClass.__doc__ = "2D Classification stage."
Class2DParticleClass.particle_sum.__doc__ = "Sum of all particles in the class. Gives a tuple with the class number first, then the particle sum."
Class2DParticleClass.reference_image.__doc__ = "Reference image."
Class2DParticleClass.class_distribution.__doc__ = (
    "Class Distribution. Proportional to the number of particles per class."
)
Class2DParticleClass.accuracy_rotations.__doc__ = "Accuracy rotations."
Class2DParticleClass.accuracy_translations_angst.__doc__ = (
    "Accuracy translations angst."
)
Class2DParticleClass.estimated_resolution.__doc__ = "Estimated resolution."
Class2DParticleClass.overall_fourier_completeness.__doc__ = (
    "Overall Fourier completeness."
)


class Class2D(JobType):
    def __eq__(self, other):
        if isinstance(other, Class2D):  # check this
            return self._basepath == other._basepath
        return False

    def __hash__(self):
        return hash(("relion._parser.Class2D", self._basepath))

    def __repr__(self):
        return f"Class2D({repr(str(self._basepath))})"

    def __str__(self):
        return f"<Class2D parser at {self._basepath}>"

    def _load_job_directory(self, jobdir):

        dfile, mfile = self._final_data_and_model(jobdir)

        sdfile = self._read_star_file(jobdir, dfile)
        smfile = self._read_star_file(jobdir, mfile)

        info_table = self._find_table_from_column_name("_rlnClassDistribution", smfile)

        class_distribution = self.parse_star_file(
            "_rlnClassDistribution", smfile, info_table
        )
        accuracy_rotations = self.parse_star_file(
            "_rlnAccuracyRotations", smfile, info_table
        )
        accuracy_translations_angst = self.parse_star_file(
            "_rlnAccuracyTranslationsAngst", smfile, info_table
        )
        estimated_resolution = self.parse_star_file(
            "_rlnEstimatedResolution", smfile, info_table
        )
        overall_fourier_completeness = self.parse_star_file(
            "_rlnOverallFourierCompleteness", smfile, info_table
        )
        reference_image = self.parse_star_file("_rlnReferenceImage", smfile, info_table)

        class_numbers = self.parse_star_file("_rlnClassNumber", sdfile, info_table)
        particle_sum = self._sum_all_particles(class_numbers)
        int_particle_sum = [(int(name), value) for name, value in particle_sum.items()]
        checked_particle_list = self._class_checker(
            sorted(int_particle_sum), len(reference_image)
        )

        particle_class_list = []
        for j in range(len(reference_image)):
            particle_class_list.append(
                Class2DParticleClass(
                    checked_particle_list[j],
                    reference_image[j],
                    float(class_distribution[j]),
                    accuracy_rotations[j],
                    accuracy_translations_angst[j],
                    estimated_resolution[j],
                    overall_fourier_completeness[j],
                )
            )
        return particle_class_list

    def _final_data_and_model(self, job_path):
        number_list = [
            entry.stem[6:9]
            for entry in (self._basepath / job_path).glob("run_it*.star")
        ]
        last_iteration_number = max(
            (int(n) for n in number_list if n.isnumeric()), default=0
        )
        if not last_iteration_number:
            raise ValueError(f"No result files found in {job_path}")
        data_file = f"run_it{last_iteration_number:03d}_data.star"
        model_file = f"run_it{last_iteration_number:03d}_model.star"
        for check_file in (
            self._basepath / job_path / data_file,
            self._basepath / job_path / model_file,
        ):
            if not check_file.exists():
                raise ValueError(f"File {check_file} missing from job directory")
        return data_file, model_file

    def _count_all(self, list):
        count = Counter(list)
        return count

    def _sum_all_particles(self, list):
        counted = self._count_all(
            list
        )  # This sorts them into classes and the number of associated particles before summing them all
        return counted
        # return sum(counted.values())

    def _class_checker(
        self, tuple_list, length
    ):  # Makes sure every class has a number of associated particles
        for i in range(1, length):
            if i not in tuple_list[i - 1]:
                tuple_list.insert(i - 1, (i, 0))
                print("No values found for class", i)
        return tuple_list

    def top_twenty(self, dictionary):
        return_dict = {}
        for item in dictionary:
            temp_list = sorted(dictionary[item], key=attrgetter("class_distribution"))[
                -20:
            ]
            temp_list.reverse()
            return_dict[item] = temp_list
        return return_dict



from collections import namedtuple
from relion._parser.jobtype import JobType

MCMicrograph = namedtuple(
    "MCMicrograph", ["micrograph_name", "total_motion", "early_motion", "late_motion"]
)

MCMicrograph.__doc__ = "Motion Correction stage."
MCMicrograph.micrograph_name.__doc__ = "Micrograph name. Useful for reference."
MCMicrograph.total_motion.__doc__ = (
    "Total motion. The amount the sample moved during exposure. Units angstrom (A)."
)
MCMicrograph.early_motion.__doc__ = "Early motion."
MCMicrograph.late_motion.__doc__ = "Late motion."


class MotionCorr(JobType):
    def __eq__(self, other):
        if isinstance(other, MotionCorr):  # check this
            return self._basepath == other._basepath
        return False

    def __hash__(self):
        return hash(("relion._parser.MotionCorr", self._basepath))

    def __repr__(self):
        return f"MotionCorr({repr(str(self._basepath))})"

    def __str__(self):
        return f"<MotionCorr parser at {self._basepath}>"

    def _load_job_directory(self, jobdir):
        file = self._read_star_file(jobdir, "corrected_micrographs.star")

        info_table = self._find_table_from_column_name("_rlnAccumMotionTotal", file)

        accum_motion_total = self.parse_star_file(
            "_rlnAccumMotionTotal", file, info_table
        )
        accum_motion_late = self.parse_star_file(
            "_rlnAccumMotionLate", file, info_table
        )
        accum_motion_early = self.parse_star_file(
            "_rlnAccumMotionEarly", file, info_table
        )
        micrograph_name = self.parse_star_file("_rlnMicrographName", file, info_table)

        micrograph_list = []
        for j in range(len(micrograph_name)):
            micrograph_list.append(
                MCMicrograph(
                    micrograph_name[j],
                    accum_motion_total[j],
                    accum_motion_early[j],
                    accum_motion_late[j],
                )
            )
        return micrograph_list

    @staticmethod
    def for_cache(mcmicrograph):
        return str(mcmicrograph.micrograph_name)



#!/usr/bin/env python3
"""
External job for running relion_reconstruct on each random half from select_and_split
External job
"""

import argparse
import os
import os.path
import subprocess
import pathlib

RELION_JOB_FAILURE_FILENAME = "RELION_JOB_EXIT_FAILURE"
RELION_JOB_SUCCESS_FILENAME = "RELION_JOB_EXIT_SUCCESS"


def run_job(project_dir, job_dir, in_dir, starin, args_list):
    parser = argparse.ArgumentParser()
    parser.add_argument("--class_number", dest="class_number")
    parser.add_argument("--mask_diameter", dest="mask_diam")
    parser.add_argument(
        "--j", dest="threads", help="Number of threads to run (ignored)"
    )
    args = parser.parse_args(args_list)
    if not starin.endswith(".star"):
        raise ValueError(
            "Input file for external ReconstructHalves job must end with .star"
        )

    split_starin = starin.split(".")
    splitname1 = ""
    splitname2 = ""
    if len(split_starin) == 2:
        splitname1 += split_starin[0] + "_split1." + split_starin[1]
        splitname2 += split_starin[0] + "_split2." + split_starin[1]
    else:
        for substr in split_starin[:-2]:
            splitname1 += substr + "."
            splitname2 += substr + "."
        splitname1 += split_starin[-2] + "_split1." + split_starin[-1]
        splitname2 += split_starin[-2] + "_split2." + split_starin[-1]

    model_half1 = f"3d_half1_model{args.class_number}.mrc"
    model_half2 = f"3d_half2_model{args.class_number}.mrc"

    command = [
        "relion_reconstruct",
        "--i",
        f"{pathlib.PurePosixPath(in_dir) / splitname1}",
        "--o",
        f"{pathlib.PurePosixPath(job_dir) / model_half1}",
        "--ctf",
        "true",
        "--mask_diameter",
        f"{args.mask_diam}",
    ]

    subprocess.run(command, check=True)

    command = [
        "relion_reconstruct",
        "--i",
        f"{pathlib.PurePosixPath(in_dir) / splitname2}",
        "--o",
        f"{pathlib.PurePosixPath(job_dir) / model_half2}",
        "--ctf",
        "true",
        "--mask_diameter",
        f"{args.mask_diam}",
    ]

    subprocess.run(command, check=True)


def main():
    """Change to the job working directory, then call run_job()"""
    parser = argparse.ArgumentParser()
    parser.add_argument("--out_dir", dest="out_dir", help="Output directory name")
    parser.add_argument(
        "--in_dir",
        dest="in_dir",
        help="Directory name where input star files are stored",
    )
    parser.add_argument("--i", dest="starin", help="Input star file name")
    parser.add_argument(
        "--in_mics",
        dest="mics_in",
        help="Input star file name (this one is just used to help Relion keep track of inputs)",
    )
    parser.add_argument(
        "--pipeline_control", help="Directory for pipeline control files"
    )
    known_args, other_args = parser.parse_known_args()
    project_dir = os.getcwd()
    os.makedirs(known_args.out_dir, exist_ok=True)
    os.chdir(known_args.out_dir)
    if os.path.isfile(RELION_JOB_FAILURE_FILENAME):
        print(
            " reconstruct_halves_external_job: Removing previous failure indicator file"
        )
        os.remove(RELION_JOB_FAILURE_FILENAME)
    if os.path.isfile(RELION_JOB_SUCCESS_FILENAME):
        print(
            " reconstruct_halves_external_job: Removing previous success indicator file"
        )
        os.remove(RELION_JOB_SUCCESS_FILENAME)
    try:
        os.chdir("../..")
        run_job(
            project_dir,
            known_args.out_dir,
            known_args.in_dir,
            known_args.starin,
            other_args,
        )
    except Exception:
        if os.getcwd() == project_dir:
            os.chdir(known_args.out_dir)
        open(RELION_JOB_FAILURE_FILENAME, "w").close()
        raise
    else:
        if os.getcwd() == project_dir:
            os.chdir(known_args.out_dir)
        open(RELION_JOB_SUCCESS_FILENAME, "w").close()


if __name__ == "__main__":
    main()



"""
Relion Python API
https://github.com/DiamondLightSource/python-relion
"""

import functools
import pathlib
from relion._parser.ctffind import CTFFind
from relion._parser.motioncorrection import MotionCorr
from relion._parser.class2D import Class2D
from relion._parser.class3D import Class3D
from relion._parser.relion_pipeline import RelionPipeline
import copy

__all__ = []
__author__ = "Diamond Light Source - Scientific Software"
__email__ = "scientificsoftware@diamond.ac.uk"
__version__ = "0.3.1"
__version_tuple__ = tuple(int(x) for x in __version__.split("."))


class Project(RelionPipeline):
    """
    Reads information from a Relion project directory and makes it available in
    a structured, object-oriented, and pythonic fashion.
    """

    def __init__(self, path):
        """
        Create an object representing a Relion project.
        :param path: A string or file system path object pointing to the root
                     directory of an existing Relion project.
        """
        super().__init__("Import/job001")
        self.basepath = pathlib.Path(path)
        if not self.basepath.is_dir():
            raise ValueError(f"path {self.basepath} is not a directory")
        try:
            self.load()
        except (FileNotFoundError, RuntimeError):
            pass
            # raise RuntimeWarning(
            #    f"Relion Project was unable to load the relion pipeline from {self.basepath}/default_pipeline.star"
            # )
        self.res = RelionResults()

    def __eq__(self, other):
        if isinstance(other, Project):
            return self.basepath == other.basepath
        return False

    def __hash__(self):
        return hash(("relion.Project", self.basepath))

    def __repr__(self):
        return f"relion.Project({repr(str(self.basepath))})"

    def __str__(self):
        return f"<relion.Project at {self.basepath}>"

    @property
    def _results_dict(self):
        resd = {
            "CtfFind": self.ctffind,
            "MotionCorr": self.motioncorrection,
            "Class2D": self.class2D,
            "Class3D": self.class3D,
        }
        return resd

    @property
    @functools.lru_cache(maxsize=1)
    def ctffind(self):
        """access the CTFFind stage of the project.
        Returns a dictionary-like object with job names as keys,
        and lists of CTFMicrograph namedtuples as values."""
        return CTFFind(self.basepath / "CtfFind")

    @property
    @functools.lru_cache(maxsize=1)
    def motioncorrection(self):
        """access the motion correction stage of the project.
        Returns a dictionary-like object with job names as keys,
        and lists of MCMicrograph namedtuples as values."""
        return MotionCorr(self.basepath / "MotionCorr")

    @property
    @functools.lru_cache(maxsize=1)
    def class2D(self):
        """access the 2D classification stage of the project.
        Returns a dictionary-like object with job names as keys,
        and lists of Class2DParticleClass namedtuples as values."""
        return Class2D(self.basepath / "Class2D")

    @property
    @functools.lru_cache(maxsize=1)
    def class3D(self):
        """access the 3D classification stage of the project.
        Returns a dictionary-like object with job names as keys,
        and lists of Class3DParticleClass namedtuples as values."""
        return Class3D(self.basepath / "Class3D")

    def origin_present(self):
        return (self.basepath / self.origin).is_dir()

    def load(self):
        self._jobs_collapsed = False
        self.load_nodes_from_star(self.basepath / "default_pipeline.star")
        self.check_job_node_statuses(self.basepath)
        self.collect_job_times(list(self.schedule_files))

    def show_job_nodes(self):
        self.load()
        super().show_job_nodes(self.basepath)

    @property
    def schedule_files(self):
        return self.basepath.glob("pipeline*.log")

    @property
    def results(self):
        # Project.motioncorrection.fget.cache_clear()
        self._clear_caches()
        res = []
        for jtnode in self:
            if jtnode.attributes["status"]:
                res_obj = self._results_dict.get(str(jtnode._path))
                if res_obj is not None:
                    res.append(
                        (
                            res_obj,
                            jtnode.attributes["job"],
                            jtnode.attributes["end_time_stamp"],
                        )
                    )
        self.res.consume(res)
        return self.res

    @property
    def current_jobs(self):
        self.load()
        currj = super().current_jobs
        if currj is None:
            return None
        else:
            for n in currj:
                n.change_name(self.basepath / n.name)
            return currj

    @staticmethod
    def _clear_caches():
        Project.motioncorrection.fget.cache_clear()
        Project.ctffind.fget.cache_clear()
        Project.class2D.fget.cache_clear()
        Project.class3D.fget.cache_clear()


class RelionResults:
    def __init__(self):
        self._results = []
        self._fresh_results = []
        self._seen_before = []
        self._fresh_called = False
        self._cache = {}

    def consume(self, results):
        self._results = [(r[0], r[1]) for r in results]
        if not self._fresh_called:
            self._fresh_results = self._results
            for r in results:
                if r[2] not in [p[1] for p in self._seen_before]:
                    self._cache[r[1]] = []
                    for single_res in r[0][r[1]]:
                        self._cache[r[1]].append(r[0].for_cache(single_res))
                if (r[1], r[2]) not in self._seen_before:
                    self._seen_before.append((r[1], r[2]))
        else:
            self._fresh_results = []
            results_copy = copy.deepcopy(results)
            for r in results_copy:
                if (r[1], r[2]) not in self._seen_before:
                    current_job_results = list(r[0][r[1]])
                    for single_res in current_job_results:
                        if self._cache.get(r[1]) is None:
                            self._cache[r[1]] = []
                        if r[0].for_cache(single_res) in self._cache[r[1]]:
                            r[0][r[1]].remove(single_res)
                        else:
                            self._cache[r[1]].append(r[0].for_cache(single_res))

                    self._fresh_results.append((r[0], r[1]))
                    self._seen_before.append((r[1], r[2]))

    def __iter__(self):
        return iter(self._results)

    @property
    def fresh(self):
        self._fresh_called = True
        return iter(self._fresh_results)



import enum
import functools
import logging
import os
import pathlib
import relion
import threading
import time
import zocalo.util.symlink
import zocalo.wrapper
from pprint import pprint
from relion.cryolo_relion_it import cryolo_relion_it, dls_options

logger = logging.getLogger("dlstbx.wrap.relion")

RelionStatus = enum.Enum("RelionStatus", "RUNNING SUCCESS FAILURE")

# to test:
# - create an empty directory
# - run following command to generate a recipe-wrapper:
#   dlstbx.find_in_ispyb -p 6844019 -f /dls_sw/apps/zocalo/live/recipes/ispyb-relion.json --recipe-pointer=2 --out=rw ; replace "/dls/m12/data/2021/cm28212-1/processed" "$(pwd)/processed" "/dls/m12/data/2021/cm28212-1/tmp" "$(pwd)/tmp" < rw > rw-local
# - run the wrapper:
#   dlstbx.wrap --recipewrapper=rw-local --wrap=relion --offline -v


class RelionWrapper(zocalo.wrapper.BaseWrapper):
    def run(self):
        assert hasattr(self, "recwrap"), "No recipewrapper object found"

        self.params = self.recwrap.recipe_step["job_parameters"]
        self.working_directory = pathlib.Path(self.params["working_directory"])
        self.results_directory = pathlib.Path(self.params["results_directory"])

        # Here we cheat. Ultimately we want to run the processing inside the
        # 'working' directory, and then copy relevant output into the 'results'
        # directory as we go along. In the first instance we will just run Relion
        # inside the 'results' directory, and ignore the 'working' directory
        # completely.
        self.working_directory = self.results_directory

        # Create working and results directory
        self.working_directory.mkdir(parents=True, exist_ok=True)
        self.results_directory.mkdir(parents=True, exist_ok=True)
        if self.params.get("create_symlink"):
            # Create symbolic link above directories
            # We want users to go to the most recent execution, as they can stop
            # and restart processing from SynchWeb. Thus we overwrite symlinks.
            zocalo.util.symlink.create_parent_symlink(
                self.working_directory,
                self.params["create_symlink"],
                overwrite_symlink=True,
            )
            zocalo.util.symlink.create_parent_symlink(
                self.results_directory,
                self.params["create_symlink"],
                overwrite_symlink=True,
            )

        # Relion needs to have a 'Movies' link inside the working directory
        # pointing to the image files
        movielink = "Movies"
        os.symlink(self.params["image_directory"], self.working_directory / movielink)
        self.params["ispyb_parameters"]["import_images"] = os.path.join(
            movielink, self.params["file_template"]
        )

        # Debug output
        pprint(self.params)

        # Start Relion
        self._relion_subthread = threading.Thread(
            target=self.start_relion, name="relion_subprocess_runner", daemon=True
        )
        self._relion_subthread.start()

        relion_prj = relion.Project(self.working_directory)

        while self._relion_subthread.is_alive() and not relion_prj.origin_present():
            time.sleep(0.5)

        relion_prj.load()

        while self._relion_subthread.is_alive() and False not in [
            n.attributes["status"] for n in relion_prj
        ]:
            time.sleep(1)

            logger.info("Looking for results")

            ispyb_command_list = []

            if pathlib.Path(self.params["stop_file"]).is_file():
                relion_prj.load()
                for job_path in relion_prj._job_nodes:
                    pathlib.Path(job_path.name / "RELION_EXIT_ABORTED").touch()

            relion_prj.load()

            # Should only return results that have not previously been sent
            for fr in relion_prj.results.fresh:
                ispyb_command_list.extend(ispyb_results(fr[0], fr[1]))
                logger.info(f"Fresh results found for {fr[1]}")

            if ispyb_command_list:
                logger.info(
                    "Sending commands like this: %s", str(ispyb_command_list[0])
                )
                self.recwrap.send_to(
                    "ispyb", {"ispyb_command_list": ispyb_command_list}
                )
                logger.info("Sent %d commands to ISPyB", len(ispyb_command_list))

        logger.info("Done.")
        success = True

        if (self.results_directory / "RUNNING_PIPELINER_PREPROCESS").is_file():
            os.remove(self.results_directory / "RUNNING_PIPELINER_PREPROCESS")

        return success

    def start_relion(self):
        print("Running RELION wrapper - stdout")
        logger.info("Running RELION wrapper - logger.info")

        for k, v in self.params["ispyb_parameters"].items():
            if v.isnumeric():
                self.params["ispyb_parameters"][k] = int(v)
            elif v.lower() == "true":
                self.params["ispyb_parameters"][k] = True
            elif v.lower() == "false":
                self.params["ispyb_parameters"][k] = False
            else:
                try:
                    self.params["ispyb_parameters"][k] = float(v)
                except ValueError:
                    pass
        pprint(self.params["ispyb_parameters"])

        # Prepare options object
        opts = cryolo_relion_it.RelionItOptions()
        opts.update_from(vars(dls_options))
        opts.update_from(self.params["ispyb_parameters"])

        # Write options to disk for a record of parameters used
        options_file = self.working_directory / cryolo_relion_it.OPTIONS_FILE
        logger.info(f"Writing all options to {options_file}")
        if os.path.isfile(options_file):
            logger.info(
                f"File {options_file} already exists; renaming old copy to {options_file}~"
            )
            os.rename(options_file, f"{options_file}~")
        with open(options_file, "w") as optfile:
            opts.print_options(optfile)

        success = False
        oldpwd = os.getcwd()
        try:
            os.chdir(self.working_directory)
            cryolo_relion_it.run_pipeline(opts)
            success = True
        except Exception as ex:
            logger.error(ex)
        finally:
            os.chdir(oldpwd)

        logger.info("Done.")
        return success

    def create_synchweb_stop_file(self):
        pathlib.Path(self.params["stop_file"]).touch()

    def get_status(self, job_path):
        relion_stop_files = [
            "RELION_JOB_EXIT_SUCCESS",
            "RELION_EXIT_FAILURE",
            "RELION_JOB_ABORT_NOW",
            "RELION_EXIT_ABORTED",
        ]
        # synchweb_stop_files = [synchweb stop files list]
        # job_finished_files = [relion job finished files]
        for item in relion_stop_files:
            if (job_path / item).is_file():  # or synchweb_stop_file exists:
                return RelionStatus.SUCCESS
            else:
                return RelionStatus.RUNNING
            # if job_finished_file exists:


@functools.singledispatch
def ispyb_results(relion_stage_object, job_string: str):
    """
    A function that takes Relion stage objects and job names (together
    representing a single job directory) and translates them into ISPyB
    service commands.
    """
    raise ValueError(f"{relion_stage_object!r} is not a known Relion object")


@ispyb_results.register(relion.CTFFind)
def _(stage_object: relion.CTFFind, job_string: str):
    logger.info("Generating ISPyB commands for %s ", job_string)
    ispyb_command_list = []
    for ctf_micrograph in stage_object[job_string]:
        ispyb_command_list.append(
            {
                "ispyb_command": "insert_ctf",
                "astigmatism": ctf_micrograph.astigmatism,
                "astigmatism_angle": ctf_micrograph.defocus_angle,
                "max_estimated_resolution": ctf_micrograph.max_resolution,
                "estimated_defocus": (
                    float(ctf_micrograph.defocus_u) + float(ctf_micrograph.defocus_v)
                )
                / 2,
                "cc_value": ctf_micrograph.fig_of_merit,
            }
        )
    return ispyb_command_list


@ispyb_results.register(relion.MotionCorr)
def _(stage_object: relion.MotionCorr, job_string: str):
    logger.info("Generating ISPyB commands for %s ", job_string)
    ispyb_command_list = []
    for motion_corr_micrograph in stage_object[job_string]:
        ispyb_command_list.append(
            {
                "ispyb_command": "insert_motion_correction",
                "micrograph_name": motion_corr_micrograph.micrograph_name,
                "total_motion": motion_corr_micrograph.total_motion,
                "early_motion": motion_corr_micrograph.early_motion,
                "late_motion": motion_corr_micrograph.late_motion,
                "average_motion_per_frame": (
                    float(motion_corr_micrograph.total_motion)
                ),  # / number of frames
            }
        )
    return ispyb_command_list


@ispyb_results.register(relion.Class2D)
def _(stage_object: relion.Class2D, job_string: str):
    logger.warning(
        "There are currently no ISPyB commands for the 2D classification stage %s ",
        job_string,
    )
    ispyb_command_list = []
    return ispyb_command_list


@ispyb_results.register(relion.Class3D)
def _(stage_object: relion.Class3D, job_string: str):
    logger.warning(
        "There are currently no ISPyB commands for the 3D classification stage %s ",
        job_string,
    )
    ispyb_command_list = []
    return ispyb_command_list



from collections import namedtuple
from relion._parser.jobtype import JobType

CTFMicrograph = namedtuple(
    "CTFMicrograph",
    [
        "micrograph_name",
        "astigmatism",
        "defocus_u",
        "defocus_v",
        "defocus_angle",
        "max_resolution",
        "fig_of_merit",
    ],
)
CTFMicrograph.__doc__ = "Contrast Transfer Function stage."
CTFMicrograph.astigmatism.__doc__ = "Estimated astigmatism. Units angstrom (A)."
CTFMicrograph.micrograph_name.__doc__ = "Micrograph name. Useful for reference."
CTFMicrograph.defocus_u.__doc__ = (
    "Averaged with Defocus V to give estimated defocus. Units angstrom (A)."
)
CTFMicrograph.defocus_v.__doc__ = (
    "Averaged with Defocus U to give estimated defocus. Units angstrom (A)."
)
CTFMicrograph.defocus_angle.__doc__ = "Estimated angle of astigmatism."
CTFMicrograph.max_resolution.__doc__ = (
    "Maximum resolution that the software can detect. Units angstrom (A)."
)
CTFMicrograph.fig_of_merit.__doc__ = (
    "Figure of merit/CC/correlation value. Confidence of the defocus estimation."
)


class CTFFind(JobType):
    def __eq__(self, other):
        if isinstance(other, CTFFind):  # check this
            return self._basepath == other._basepath
        return False

    def __hash__(self):
        return hash(("relion._parser.CTFFind", self._basepath))

    def __repr__(self):
        return f"CTFFind({repr(str(self._basepath))})"

    def __str__(self):
        return f"<CTFFind parser at {self._basepath}>"

    @property
    def job_number(self):
        jobs = [x.name for x in self._basepath.iterdir()]
        return jobs

    def _load_job_directory(self, jobdir):
        file = self._read_star_file(jobdir, "micrographs_ctf.star")

        info_table = self._find_table_from_column_name("_rlnCtfAstigmatism", file)

        astigmatism = self.parse_star_file("_rlnCtfAstigmatism", file, info_table)
        defocus_u = self.parse_star_file("_rlnDefocusU", file, info_table)
        defocus_v = self.parse_star_file("_rlnDefocusV", file, info_table)
        defocus_angle = self.parse_star_file("_rlnDefocusAngle", file, info_table)
        max_resolution = self.parse_star_file("_rlnCtfMaxResolution", file, info_table)
        fig_of_merit = self.parse_star_file("_rlnCtfFigureOfMerit", file, info_table)

        micrograph_name = self.parse_star_file("_rlnMicrographName", file, info_table)

        micrograph_list = []
        for j in range(len(micrograph_name)):
            micrograph_list.append(
                CTFMicrograph(
                    micrograph_name[j],
                    astigmatism[j],
                    defocus_u[j],
                    defocus_v[j],
                    defocus_angle[j],
                    max_resolution[j],
                    fig_of_merit[j],
                )
            )
        return micrograph_list

    @staticmethod
    def for_cache(ctfmicrograph):
        return str(ctfmicrograph.micrograph_name)



import collections.abc
import os
from gemmi import cif


class JobType(collections.abc.Mapping):
    def __eq__(self, other):
        if isinstance(other, JobType):  # check this
            return self._basepath == other._basepath
        return False

    def __hash__(self):
        return hash(("relion._parser.JobType", self._basepath))

    def __init__(self, path):
        self._basepath = path
        self._jobcache = {}

    def __iter__(self):
        return iter(self.jobs)

    def __len__(self):
        return len(self.jobs)

    def __repr__(self):
        return f"JobType({repr(str(self._basepath))})"

    def __str__(self):
        return f"<JobType parser at {self._basepath}>"

    @property
    def jobs(self):
        return sorted(
            d.name
            for d in self._basepath.iterdir()
            if d.is_dir() and not d.is_symlink()
        )

    def __getitem__(self, key):
        if not isinstance(key, str):
            raise KeyError(f"Invalid argument {key!r}, expected string")
        if key not in self._jobcache:
            job_path = self._basepath / key
            if not job_path.is_dir():
                raise KeyError(
                    f"no job directory present for {key} in {self._basepath}"
                )
            self._jobcache[key] = self._load_job_directory(key)
        return self._jobcache[key]

    def _load_job_directory(self, jobdir):
        raise NotImplementedError("Load job directory not implemented")

    def _read_star_file(self, job_num, file_name):
        full_path = self._basepath / job_num / file_name
        gemmi_readable_path = os.fspath(full_path)
        star_doc = cif.read_file(gemmi_readable_path)
        return star_doc

    def parse_star_file(self, loop_name, star_doc, block_number):
        data_block = star_doc[block_number]
        values = data_block.find_loop(loop_name)
        values_list = list(values)
        if not values_list:
            print("Warning - no values found for", loop_name)
        return values_list

    def _find_table_from_column_name(self, cname, star_doc):
        for block_index, block in enumerate(star_doc):
            if list(block.find_loop(cname)):
                return block_index
        return None

    @staticmethod
    def for_cache(element):
        return None



import pathlib
import unicodedata

import workflows.recipe
from workflows.services.common_service import CommonService


class RelionStopService(CommonService):
    """A service that creates 'stop' files that interrupt Relion processing jobs."""

    # Human readable service name
    _service_name = "Relion Stop Service"

    # Logger name
    _logger_name = "relion.zocalo.service"

    def initializing(self):
        """Subscribe to the Relion Stop queue.
        Received messages must be acknowledged."""
        self.log.debug("Relion Stop service starting")
        workflows.recipe.wrap_subscribe(
            self._transport,
            "relion.dev.stop",
            self.receive_msg,
            acknowledgement=True,
            log_extender=self.extend_log,
        )

    def receive_msg(self, rw, header, message):
        """Stop a Relion processing job."""
        parameters = rw.recipe_step["parameters"]
        stop_file = parameters.get("stop_file")
        if not stop_file:
            self.log.error("No stop file defined in message")
            self._transport.nack(header)
            return
        stop_file = pathlib.Path(
            "".join(
                ch for ch in stop_file if not unicodedata.category(ch).startswith("C")
            )
        )
        if not stop_file.is_absolute():
            self.log.error(f"Stop file path '{stop_file}' is not absolute")
            self._transport.nack(header)
            return
        if not stop_file.parent.is_dir():
            self.log.error(f"Parent of stop file path '{stop_file}' does not exist")
            self._transport.nack(header)
            return
        if stop_file.exists():
            self.log.info(f"Requested stop file '{stop_file}' already exists")
            self._transport.ack(header)
            return
        self.log.info(f"Creating stop file '{stop_file}'")
        pathlib.Path(stop_file).touch()
        self._transport.ack(header)
